/* convert a structure factor file from one format to another */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include "rcsb_math.h"
#include "sf_convert.h"
#include "util.h"
#include "TableFile.h"
#include "CifFile.h"
#include "CifParserBase.h"

// externals 
const char *CCP4, *URL_USERS_DATA, *USERS_DATA, *CGI_PATH, *PATHTMP;  
char SYMM[20],AUDIT[512],EXECUTE[512],FREERV[20],PDBID[80],INFO[512];
int TEST=0, SINGLE=0, FORMAT=0, DATA_SET=1, DATA_PHS=0;
int RCSB=0, ADD_MAP=0, CIF_CHECK=0, VERB=0;
float CELL[6], WAVE=0, RESCUT=0, SIGCUT=0;
FILE *FTMP1=NULL, *FTMP2=NULL;

void get_sf_info(char *diagfile);

int main(int argc, char **argv)
{
    char inpfile[512], outfile[512], diagfile[512],outfile_tmp[512];
    char format_inp[20], format_out[20], label_items[512], pdbfile[512];
    char cgi_path[512], I_F_type[10], *temp;
    int i, j,mtz_cgi=0, cns_cgi=0, label=0,iscif=0;
    float flag=0;
    

    FTMP1=tmpfile();
    FTMP2=tmpfile();
    RCSB=0; //give 1 to source the CCP4 in the CGI web (only 4 rcsb)
    

//    PATHTMP = "/tmp/";  //specify the path for computing
    PATHTMP = "./";  //specify the path for computing
    strcpy(PDBID, "xxxx");
    strcpy(FREERV, "");
    strcpy(AUDIT, "");
    strcpy(label_items, "");
    strcpy(inpfile, "");
    strcpy(outfile, "");
    strcpy(diagfile, "");
    strcpy(pdbfile, "");
    strcpy(I_F_type, "");
    strcpy(format_inp, "");
   
    check_input_error(argc, argv);

    printf("-----------------------------------------------------------------\n");
    printf("|           sf_convert (version: %s)                         |\n",  VERSION );
    printf("-----------------------------------------------------------------\n");

    
    CCP4 = getenv("CCP4");
    
    if((temp = getenv("PROG_VARI")) != NULL) { //only for CGI related
        sprintf(EXECUTE,"%s/bin/sf_convert", temp);
    }else{
        sprintf(EXECUTE,"sf_convert");
    }
    
    for(i=0; i<6; i++){CELL[i]=0;}
    WAVE=0;

    for(i=0; i<argc; i++){
        if(!strcmp_case(argv[i], "-i")){
            check_argv(argv[i], i, argc);
            strcpy(format_inp, argv[i+1]);
            
            
        }else if(!strcmp_case(argv[i], "-o")){
            check_argv(argv[i], i, argc);
            strcpy(format_out, argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-sf")){
            check_argv(argv[i], i, argc);
            file_handle(argv[i+1], inpfile);
            check_inpfile_name(inpfile, PDBID);
            
        }else if(!strcmp_case(argv[i], "-out")){
            check_argv(argv[i], i, argc);
            strcpy(outfile, argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-diags")){
            check_argv(argv[i], i, argc);
            strcpy(diagfile, argv[i+1]);

        }else if(!strcmp_case(argv[i], "-rescut")){ //cut resolution  
            check_argv(argv[i], i, argc);
            RESCUT= atof(argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-sigcut")){ //cut sigma (I/SigI)
            check_argv(argv[i], i, argc);
            SIGCUT= atof(argv[i+1]);

        }else if(!strcmp_case(argv[i], "-freer")){
            check_argv(argv[i], i, argc);
            strcpy(FREERV, argv[i+1]);
            sprintf(INFO,"Note: %s is used for free data set.", FREERV);
            pinfo(INFO,0);
            

        }else if(!strcmp_case(argv[i], "-cif2pdb")){
            sprintf(pdbfile,"%s.PDB", argv[i+1]);
            cif2pdb(argv[i+1], pdbfile);
            exit(0);
        }else if(!strcmp_case(argv[i], "-pdb")){
            check_argv(argv[i], i, argc);
            if (argv[i+1][0] == '-') continue;
            
            strcpy(pdbfile, argv[i+1]);
            if (is_cif(argv[i+1])>3){ // cif file convert to pdb
                sprintf(pdbfile,"%s.pdb", argv[i+1]);
                cif2pdb(argv[i+1], pdbfile);
                iscif=1;
            }
            get_cell_pdb(pdbfile, PDBID);
            
            
        }else if(!strcmp_case(argv[i], "-man")||!strcmp_case(argv[i], "-cell")){
            get_cell_man(argv, argc, i, &j);
            i=j;
            
        }else if(!strcmp_case(argv[i], "-flag")){
            check_argv(argv[i], i, argc);
            if(argv[i+1][0] != '-'){
                sscanf(argv[i+1], "%f", &flag);
                if(flag>1)
                    TEST=int(100/flag + 0.5);
                else{
                    TEST=int(100/5 + 0.5);  // 5%
                    printf("The flag is < 1%%, 5%% has been assigned\n");
                }
            }
            
            sprintf(INFO,"Note: the free data set is manually assigned as %s%%.", argv[i+1]);
            pinfo(INFO,0);
            
        }else if(!strcmp_case(argv[i], "-sf_type")){ //I or F
            check_argv(argv[i], i, argc);
            strcpy(I_F_type, argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-pdb_id")){
            check_argv(argv[i], i, argc);
            sprintf(PDBID, "%s", argv[i+1]);
            lowerstr(PDBID);
            
        }else if(!strcmp_case(argv[i], "-label")){
            check_argv(argv[i], i, argc);
            for(j=i+1; j<argc;j++){
                if(argv[j][0]=='-'){
                    i=j-1;
                    break;
                }
                strcat(label_items, argv[j]);
                strcat(label_items, " ");
            }
            label=1;
            if (strlen(label_items)<5){
                sprintf(INFO,"Error: Please add the correct label names, then try again!");
                pinfo(INFO,0);
                return 0;
            }
            
            
        }else if(!strcmp_case(argv[i], "-parse")){
            check_argv(argv[i], i, argc);
            if(i<argc-1) cifparse(argv[i+1],argv[i+1], 80 );
            return 0;
            
        }else if(!strcmp_case(argv[i], "-valid")){ //only fo validation
            check_argv(argv[i], i, argc);
            validate_sf_format(argv[i+1]);
            return 0;
            
        }else if(!strcmp_case(argv[i], "-remtz")){ //reformat mtz format
            check_argv(argv[i], i, argc);
            reformat_mtz_output(argv[i+1],PDBID);
            return 0;
            
        }else if(!strcmp_case(argv[i], "-tmp_path")) { //for CGI only
            check_argv(argv[i], i, argc);
            PATHTMP = argv[i+1];
            
        }else if(!strcmp_case(argv[i], "-path")){  //for CGI only
            check_argv(argv[i], i, argc);
            strcpy(cgi_path, argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-data_phs")){ //only for CGI
            DATA_PHS=1;
            
        }else if(!strcmp_case(argv[i], "-mtz_man_html")){ //only for CGI
            check_argv(argv[i], i, argc);
            mtz_cgi =1;
            if(atoi(argv[i+1])>=1) DATA_SET=atoi(argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-cns_man_html")){ //only for CGI
            check_argv(argv[i], i, argc);
            cns_cgi =1;
            if(atoi(argv[i+1])>=1) DATA_SET=atoi(argv[i+1]);
            
        }else if(!strcmp_case(argv[i], "-url_users_data")){ //only for CGI
            check_argv(argv[i], i, argc);
            URL_USERS_DATA=argv[i+1];
            
        }else if(!strcmp_case(argv[i], "-users_data")){ //only for CGI
            check_argv(argv[i], i, argc);
            USERS_DATA=argv[i+1];
            
        }else if(!strcmp_case(argv[i], "-cgi_bin")){ //only for CGI
            check_argv(argv[i], i, argc);
            CGI_PATH=argv[i+1];
            if (CGI_PATH[0]=='-') CGI_PATH= "";
            
            
        }else if(!strcmp_case(argv[i], "-correct_list")){
            check_argv(argv[i], i, argc);
            correct_list(argv[i+1]);
//            correct_list_file(argv[i+1], argv[i+2]);
//first file should be the annotated pdbid list, second file has 3 columns
//             
            return 0;
            
        }else if(!strcmp_case(argv[i], "-correct")){ //Correct cif token errors
            check_argv(argv[i], i, argc);
            char **line; int nline; FILE *fw=NULL;
            sprintf(outfile, "%s_new", argv[i+1]);
            line=get_lines_from_file(argv[i+1], &nline);
            rename_cif_token(nline, line, outfile);
            printf("The corrected cif file =%s\n", outfile);
            fw=fopen(outfile,"w");
            for(j=0; j<nline; j++){ fprintf(fw, "%s",line[j]);}
            fclose(fw);
            return 0;
            
        }else if(!strcmp_case(argv[i], "-audit")){
//update audit: -audit date :correction@
//(date format: 2003-11-25; correction start with : and end with @)
            for(j=i+1; j<argc; j++){
                strcat(AUDIT, argv[j]);
                strcat(AUDIT, " ");
                
                if(strstr(argv[j], "@")){
                    i=j;
                    break;
                }
            }
        }else if(!strcmp_case(argv[i], "-verb")){
            VERB=1;
        }
    }

    
    
    for(i=0; i<argc; i++){ //no arguments needed
        if(!strcmp_case(argv[i], "-add_map")) { //extract map coefficients
            ADD_MAP=1;
            
        }else if(!strcmp_case(argv[i], "-cif_check")){
            CIF_CHECK=1; //check pdbx dic by cif_check
            
        }else if(!strcmp_case(argv[i], "-single")){
            SINGLE=1;
            
        }
    }
    
    for(i=0; i<argc; i++){
        if(!strcmp_case(argv[i], "-reformat")){ //correct SF header

            sprintf(outfile, "%s_new", argv[i+1]);
            reformat_sfhead(argv[i+1], outfile);
            if(strlen(pdbfile)>0)  make_complete_SF(pdbfile, outfile);

            return 0;
            
        }else if(!strcmp_case(argv[i], "-format")){ //guess sf format
            guess_sf_format(inpfile, format_inp);
            return 0;
        }
    }

    
    if(!strlen(format_inp) && strlen(format_out) ){//No input format -i
        guess_sf_format(inpfile,format_inp);
        printf("If the guessed input format (%s) is wrong, use command below:\n",
               format_inp);
        printf("sf_convert  -i input_format -o output_format -sf data_file\n");
    }
    
    if(!strcmp_case(format_inp, "MMCIF")){
        if(is_cif(inpfile)<4){
            sprintf(INFO,"Error! The input file (%s) is not a mmCIF format!", inpfile);
            pinfo(INFO,0);
            return 0;
        }
    }
    
    if(!strcmp_case(format_inp, "MTZ")||!strcmp_case(format_out, "MTZ")){
        
        if(!strcmp_case(format_inp, "MTZ") && !is_mtz(inpfile)){
            sprintf(INFO,"Error! The input file (%s) is not a MTZ format!", inpfile);
            pinfo(INFO,0);
            return 0;
        }
            
        if((CCP4=getenv("CCP4"))==NULL ){
            sprintf(INFO, "Error! The CCP4 environment is not set.\n");
            pinfo(INFO, 0);
            printf("Please source the file 'ccp4.setup' in ccp4-?/include/ and try again.\n");
            exit(0);
        }
    }
    
    if(mtz_cgi>0){ //only for CGI
        get_mtz_infor(inpfile, cgi_path, PDBID);
        exit(0);
    }
    
    if(cns_cgi>0){ //only for CGI
        get_cns_html(inpfile, cgi_path, PDBID);
        printf("The HTML file is %scns_man_convert.html\n",PATHTMP);
        exit(0);
    }

    printf("\nInput File Name = %s : (format=%s)\n",inpfile, format_inp);
    
    if(!strcmp_case(format_inp, "mmcif") && is_cif(inpfile)<3){
        printf("Error! The input file (%s) is not a mmcif format!\n", inpfile);
        exit(0);
    }
    
//  ================ converting start  ===========================    
    if(strlen(outfile)<=0) {sprintf(outfile, "%s.%s",inpfile, format_out);}
    sprintf(outfile_tmp, "%s_new", inpfile);
    
    if(label>0){
//convert CNS/MTZ to mmCIF for irregular labels or multiple data sets.
        printf("Manual conversion from %s to mmCIF format ...\n", format_inp);
        
        if(!strcmp_case(format_inp, "CNS")||!strcmp_case(format_inp, "Xplor"))
            cns_multi_sets(inpfile, outfile, label_items, PDBID);
        else if(!strcmp_case(format_inp, "MTZ"))
            mtz_man_sets(inpfile, outfile, label_items, PDBID);
        
        reformat_sfhead(outfile, outfile_tmp);
        if(strlen(pdbfile)>0) make_complete_SF(pdbfile, outfile_tmp);
        cif2cif_sf_all(outfile_tmp, outfile, PDBID, 1);
        
    }else  if(!strcmp_case(format_inp, "mmcif") && !strcmp_case(format_out, "mmcif")){
//mmcif to mmcif conversion:
        reformat_sfhead(inpfile, outfile_tmp);
        
        if(strlen(pdbfile)>0) make_complete_SF(pdbfile, outfile_tmp);
        cif2cif_sf_all(outfile_tmp, outfile, PDBID, 1);
        if (VERB==0) delete_file(outfile_tmp);
        
    }else if(strcmp_case(format_inp, "mmcif") && !strcmp_case(format_out, "mmcif")){
//non-mmcif to mmcif conversion:
        
        various_2_cif(inpfile, outfile, format_inp, PDBID, I_F_type);
        
        reformat_sfhead(outfile, outfile_tmp);
        
        if(strlen(pdbfile)>0) make_complete_SF(pdbfile, outfile_tmp);
       
        
        cif2cif_sf_all(outfile_tmp, outfile, PDBID, 1);
        add_audit(outfile);
        if (VERB==0) delete_file(outfile_tmp);
        
    }else if(!strcmp_case(format_inp, "mmcif") && strcmp_case(format_out, "mmcif")){
//mmcif to non-mmcif conversion:        
        cif_2_various(inpfile, pdbfile, outfile, format_out);
        
    }else {
//non-mmcif to non-mmcif conversion: (bridged by mmcif)    
        various_2_cif(inpfile, outfile_tmp, format_inp, PDBID, I_F_type);
        cif_2_various(outfile_tmp, pdbfile, outfile, format_out);
        if (VERB==0) delete_file(outfile_tmp);
    }
    
    get_sf_info(diagfile);
    
    if (iscif && VERB==0) delete_file(pdbfile);
    
    printf("\nOutput File Name = %s : (%s format)\n\n", outfile, format_out);
}

//===============================================================================


void get_sf_info(char *diagfile)
{
    char str[2000];
    FILE *fw=NULL;
    

    if (strlen(diagfile)){  //for D&N only
        int m=0;
        FILE *fw1=NULL;
        
        rewind(FTMP1);
        fw1=fopen(diagfile, "w");
        while(fgets(str, sizeof str, FTMP1) != NULL){
            
            rid_of_front_end_space(str);
            if (strlen(str)==0) continue;
            m=m+1;
            fprintf(fw1,"%s\n", str);
        }
        if (m==0){
            fprintf(fw1,"No Error/Warning messages were found.\n");
        }
        fclose(fw1);
    }
    
 
    delete_file("sf_information.cif");
    fw =fopen("sf_information.cif", "w");
    rewind(FTMP1);
    rewind(FTMP2);
    fprintf(fw, "data_info\n\n");
    fprintf(fw, "_sf_convert.error\n;\n");
    while(fgets(str, sizeof str, FTMP1) != NULL){
        fprintf(fw,"%s", str);
    }
    fprintf(fw, ";\n\n");

    fprintf(fw, "_sf_convert.sf_information\n;\n");
    while(fgets(str, sizeof str, FTMP2) != NULL){
        fprintf(fw,"%s", str);
    }
    fprintf(fw, ";\n");

    if (FTMP1) {fclose(FTMP1);}
    if (FTMP2) {fclose(FTMP2);}
    
    fclose(fw);
    
    printf("The SF information is written in file=sf_information.cif\n");

    
}
        
void check_argv(char *arg1, int i, int argc)
{
    
//    printf("checking %s %d %d\n", arg1, i, argc);
    
    if (i>=argc-1){
        sprintf(INFO,"Error: A value must be given after the argument %s !", arg1);
        pinfo(INFO,0);
        exit(0);
    }
}
    
void validate_sf_format(char *inpfile)
//validate the SF format,detect any errors in the file.
{
    string diags, blockId;
    const char *Bid;
    char str[512],tmp[512], *cel, *sym;
    char **H, **K, **L, **status, **fs1, **fs2, **is1,**is2, **fom, **phase ;
    char **ips,**fps, **f,  **f2,  **in, **fm, **fp, **im,**ip, **fu;
    int k1,k2,k3,no=0,nf=0,ns=0, nnum=0,nob=0, nfom=0, nph=0 ;
    long n1=0, i;
    FILE  *finp=NULL;

    strcpy(tmp, "");
    precheck_sf(inpfile, tmp);
    
    CifFile * fobjR = NULL;
    
    CifParser* cifParserR = NULL;
    fobjR = new CifFile(true);
    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
    cifParserR->Parse(inpfile, diags);
    delete(cifParserR);

    vector<string> blockNames;
    fobjR->GetBlockNames(blockNames);
    blockId = fobjR->GetFirstBlockName();
    
    Bid=blockId.c_str();

    H  = get_attribute_values(fobjR, Bid, "refln", "index_h", &n1);
    K  = get_attribute_values(fobjR, Bid, "refln", "index_k", &n1);
    L  = get_attribute_values(fobjR, Bid, "refln", "index_l", &n1);

    if(!(H && K && L)){
        sprintf(INFO, "Error! File (%s) has no HKL. (do not check further)", inpfile);
        pinfo(INFO, 0);
        return;
    }
    
    cel  = get_attribute_value(fobjR, Bid, "cell", "length_a");
    sym  = get_attribute_value(fobjR, Bid, "symmetry", "space_group_name_H-M");
    if (!cel || !sym){
        sprintf(INFO, "Warning: File (%s) has no cell/symmetry.", inpfile);
        pinfo(INFO, 0);
    }
    
    
    fu= get_attribute_values(fobjR, Bid, "refln", "F_meas", &n1);
    fs2 = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma", &n1);
    
    f = get_attribute_values(fobjR, Bid, "refln", "F_meas_au", &n1);
    fs1 = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_au", &n1);
    
    f2= get_attribute_values(fobjR, Bid, "refln", "F_squared_meas", &n1);
    is2 = get_attribute_values(fobjR, Bid, "refln", "F_squared_sigma", &n1);
    
    in= get_attribute_values(fobjR, Bid, "refln", "intensity_meas", &n1);
    is1 = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma", &n1);

    fm= get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus", &n1);
    fp= get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus", &n1);
    fps = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus_sigma", &n1);

    im= get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus", &n1);
    ip= get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus", &n1);
    ips = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus_sigma", &n1);

    status  = get_attribute_values(fobjR, Bid, "refln", "status", &n1);
    fom  = get_attribute_values(fobjR, Bid, "refln", "fom", &n1);
    phase  = get_attribute_values(fobjR, Bid, "refln", "phase_calc", &n1);
    if (!phase) phase  = get_attribute_values(fobjR, Bid, "refln", "phase_meas", &n1);
       
    if(n1<5){
        sprintf(INFO, "Error! File (%s) not be converted. (check format, or use manual mode)", inpfile);
        pinfo(INFO, 0);
    }

    if(!(H || K || L)){
        sprintf(INFO, "Error: File (%s) has no H, K, L index.(check SF type)", inpfile);
        pinfo(INFO, 0);
    }
    
    if(!(fu || f || f2 || in || fp || fm || ip ||im)){
        sprintf(INFO,"Error! File (%s) has no madatory items (F|I|F2|F+|F-|I+|I-)", inpfile);
        pinfo(INFO, 0);
    }
    
    if(fu){
        sprintf(INFO, "Warning! File (%s) is labeled F as _refln.F_meas (use _refln.F_meas_au).", inpfile);
        pinfo(INFO, 0);
    }

    if(!fs1 && !fs2 && !is1 && !is2 && !fps && !ips ){
        sprintf(INFO, "Warning! File (%s) has no sigma values.",inpfile );
        pinfo(INFO, 0);
    }
    
    if(fs1){
        ns=0;
        for(i=0; i<n1; i++){
            if(atof(fs1[i])>0.0001) ns++;
        }
        if(ns<1) {
            sprintf(INFO, "Warning! File (%s) has all zero Sigma values (SigF).", inpfile);
            pinfo(INFO, 0);
        }
    }
    if(is1){
        ns=0;
        for(i=0; i<n1; i++){
            if(atof(is1[i])>0.0001) ns++;
        }
        if(ns<1) {
            sprintf(INFO, "Warning! File (%s) has all zero Sigma values (SigI).", inpfile);
            pinfo(INFO, 0);
        }
    }
    
    if(fps){
        ns=0;
        for(i=0; i<n1; i++){
            if(atof(fps[i])>0.0001) ns++;
        }
        if(ns<1) {
            sprintf(INFO, "Warning! File (%s) has all zero Sigma values (SigF+)", inpfile);
            pinfo(INFO, 0);
        }
    }
    
    no=0; nf=0;
        
    for(i=0; i<n1; i++){
        k1=atoi(H[i]);
        k2=atoi(K[i]);
        k3=atoi(L[i]);
      
        if (fom && fabs(atof(fom[i])) > 1.01) {
            nfom++;
            if(nfom==1){
                sprintf(INFO,"Error: File (%s) has wrong values of FOM (%s).", inpfile, fom[i]);
                pinfo(INFO, 0);
            }
        }
        
        if(phase && fabs(atof(phase[i]))>361.0){
            nph++;
            if(nfom==1){
                sprintf(INFO,"Error: File (%s) has wrong values of phase (%s).", inpfile, phase[i]);
                pinfo(INFO, 0);
            }
        }
        
        
        if(k1>1000 || k1<-1000 || k2>1000 || k2<-1000 || k3>1000 || k3<-1000){
            nf++;
            if(nf==1){
                sprintf(INFO, "Error! File (%s) has H K L (%4d %4d %4d) too large (check!).",inpfile, k1,k2,k3);
                pinfo(INFO, 0);
            }
        }else if((fu && atof(fu[i])<0) || (f && atof(f[i])<0) ||
                 (fp && atof(fp[i])<0) || (fm && atof(fm[i])<0)){
            no++;
            if(no==1){
                sprintf(INFO, "Warning! File (%s) has negative amplitudes (F)! (need check).", inpfile);
                pinfo(INFO, 0);
            }
        }
    }
    
    no=0; nf=0; ns=0;
    if(status){
        for(i=0; i<n1; i++){
            strcpy(tmp,status[i]);
            upperstr(tmp);
            if(tmp[0]=='X' || tmp[0]=='<' ||  tmp[0]=='H' ||  tmp[0]=='L' || tmp[0]=='.' ||
               (tmp[0]=='-' and strlen(tmp)==1)) {
                nob++;  //non-observed
            }else if(status[i][0] == 'o'){
                no++;
            }else if (status[i][0] == 'f'){
                nf++;
            }else  //assigned as number
                nnum++;
        }
        
        if(nf>no){
            sprintf(INFO, "Error! File (%s) has status more 'f' than 'o'.", inpfile );
            pinfo(INFO, 0);
        }
        
        if(nnum >100){
            sprintf(INFO, "Error! File (%s) has status assigned to numbers.", inpfile );
            pinfo(INFO, 0);
        }
        
        if(n1-nob<10 && n1>1000){
            sprintf(INFO, "Error! File (%s) has status ALL assigned as (X, <, H, L, -).", inpfile );
            pinfo(INFO, 0);
        }
        
    }else{
        sprintf(INFO, "Warning: File (%s) has no status (free set) assigned.", inpfile );
        pinfo(INFO, 0);
    }
        
    sprintf(tmp,"%s-parser.log", inpfile);
    if((finp=fopen(tmp, "r"))!=NULL) {
        while(fgets(str, sizeof str, finp) != NULL){
            if(strstr_case(str, "error") || strstr_case(str, "warn")){
                rid_of_front_end_space(str);
                sprintf(INFO, "%s (file=%s)", str, inpfile);
                pinfo(INFO, 0);
            }
        }
        fclose(finp);
    }

}

void check_inpfile_name(char *file, char *pdbid)
//if file name is r(id)sf.org or /?/r(id)sf.org, take id as entry id
{
    char str[512];
    
    strcpy(str, file);
    if (strchr(file, '/')) {
        strcpy(str, strrchr(file, '/') +1);
    }
    
   
    if (strlen(str)==11 && str[0]=='r' && !strncmp(str+5, "sf.org",6)){
        strncpy(pdbid,str+1,4);
        pdbid[4]='\0';
    }

}

void check_data_block_name(char *file, char *block_id, char *pdbid, int num)
// num=0 is the first data block
{
    int n=0, prob=0;


    rid_of_front_end_space(block_id);
    if(strchr(block_id, '/')) replace_char(block_id, '/', '_');
    if(strchr(block_id, ')')) replace_char(block_id, ')', '_');
    if(strchr(block_id, '(')) replace_char(block_id, '(', '_');

    if(strchr(block_id, '/') || strchr(block_id, ')') || strchr(block_id, '(')){
        sprintf(INFO, "Note: File (%s) has blockID (%d: %s) not standard(r????sf, r????Asf ...).", file, num, block_id);
        pinfo(INFO,0);
        return;
    }
    
    n=strlen(block_id);
    
    if (num==0){
        if (n != 7 || block_id[0] !='r' || !strstr(block_id +5, "sf")){
            prob=1;
        }else{
            strncpy(pdbid, block_id + 1, 4);
            pdbid[4]='\0';
        }
        
    }else {
        if (n != 8 || block_id[0] !='r' || !strstr(block_id +6, "sf")){
            prob=1;
        }else if(num<25 && (block_id[5]<65 || block_id[5]>90)){
            prob=1;
        }else{
            strncpy(pdbid, block_id + 1, 4);
            pdbid[4]='\0';
        }
    }
    if (prob==1){
        sprintf(INFO, "Note: File (%s) has blockID (%d: %s) not standard(r????sf, r????Asf ...).", file, num, block_id);
        pinfo(INFO,0);
    }
    rid_of_space(pdbid);
    
}


void make_complete_SF(char *pdbfile, char *outfile)
// add categories to SF file, if not exist
{
    char **line1=NULL, **line2=NULL, date[40],pdbid[10], symm[20], tmp[40];
    int i, nline1, nline2, n1=0,n2=0,n3=0,n4=0,n5=0, nn;
    float cell[6], wave=0;
    FILE *fo;

    
    
    line1=get_lines_from_file(pdbfile, &nline1);
    line2=get_lines_from_file(outfile, &nline2);
    
    fo=fopen(outfile, "w");

    for(i=0; i<6; i++)cell[i]=0;
    strcpy(date, "?");
    strcpy(symm, "?");
    strcpy(pdbid, "xxxx");
    for(i=0; i<nline1; i++){ //PDB infor.
        if(!strncmp(line1[i], "ATOM  ", 6) || !strncmp(line1[i], "HETATM", 6)) break;
        if (!strncmp(line1[i],"HEADER ", 7) && strlen(line1[i])>62 ){
            sscanf(line1[i]+61, "%s",  tmp);
            if(strlen(tmp)>0){
                lowerstr(tmp);
                strcpy(pdbid, tmp);
            }   
        }else if(!strncmp(line1[i],"CRYST1 ", 7)){
            sscanf(line1[i]+7, "%f%f %f%f %f%f ",
                   &cell[0], &cell[1], &cell[2], &cell[3], &cell[4], &cell[5]);
            get_string(line1[i],54,66,symm );
            rid_of_front_end_space(symm);
            
        }else if(strstr(line1[i],"REMARK 200  WAVELENGTH OR RANGE        (A) :")){
            sscanf(strchr(line1[i], ':')+1,"%s", tmp);
            if(isdigit(tmp[0])) wave=atof(tmp);
            
        }
    }
    if(line1 && nline1)free_memory_2d(line1, nline1);
    
    
    for(i=0; i<nline2; i++){ //SF info
        if (strchr(line2[i], '_')) rid_of_front_space(line2[i]);
        if(!strncmp(line2[i], "_audit.", 7)){
            n1=1;
        }else if (!strncmp(line2[i], "_cell.length_a", 14)){
            n2=1;
        }else if (!strncmp(line2[i], "_symmetry.space_group_name_H-M", 30)){
            n3=1;
        }else if (!strncmp(line2[i], "_diffrn_radiation.", 18)){
            n4=1;
        }else if (strstr(line2[i], "_diffrn_radiation_wavelength.wavelength")){
            n5=1;
        }else if(i>300 && !strncmp(line2[i],"data_",5)) break; 
    }

    
    nn=0;
    for(i=0; i<nline2; i++){ //write SF info
        if(strstr(line2[i], "data_mtz2cif") || strstr(line2[i], "data_rxxxxsf") ){
            fprintf(fo, "data_r%ssf", pdbid);
        }else{
            fprintf(fo, "%s", line2[i]);
        }
        
        if(i< nline2-2 && !strncmp(line2[i+1],"loop_",5) &&  !strncmp(line2[i+2],"_refln.", 7) ) {
            
            nn++;
            if(nn==1 && n1==0) {
                fprintf(fo,"#\n");
                fprintf(fo,"_audit.revision_id      1_0\n");
                fprintf(fo,"_audit.creation_date  '%s'\n", date);
                fprintf(fo,"_audit.update_record    'Initial release'\n#\n");
                sprintf(INFO, "Note: file (%s) has no _audit. (auto added)",outfile);
                pinfo(INFO,0);
            }
            if(nn==1 && n2==0) {
                fprintf(fo, "#\n_cell.entry_id    %s\n", pdbid);
                fprintf(fo, "_cell.length_a  %.3f\n", cell[0] );
                fprintf(fo, "_cell.length_b  %.3f\n", cell[1]);
                fprintf(fo, "_cell.length_c  %.3f\n", cell[2]);
                fprintf(fo, "_cell.angle_alpha  %.3f\n", cell[3]);
                fprintf(fo, "_cell.angle_beta  %.3f\n", cell[4]);
                fprintf(fo, "_cell.angle_gamma  %.3f\n#\n", cell[5]);
                
                sprintf(INFO, "Note: file (%s) has no _cell. (auto added)",outfile);
                pinfo(INFO,0);
            }
            if(nn==1 && n3==0) {
                fprintf(fo, "#\n_symmetry.entry_id    %s\n", pdbid);
                fprintf(fo, "_symmetry.space_group_name_H-M '%s'\n#\n", symm);
                
                sprintf(INFO, "Note: file (%s) has no _symmetry. (auto added)",outfile);
                pinfo(INFO,0);

            }
            
            if(nn==1 && n4==0 && n5==0) {
                if(wave>0.001){
                    
                    fprintf(fo, "#\n_diffrn_radiation_wavelength.id  1\n");
                    fprintf(fo, "_diffrn_radiation_wavelength.wavelength    %.5f\n#\n", wave);
                    
                    sprintf(INFO, "Note: file (%s) has no _diffrn_radiation_wavelength.(auto added)",outfile);
                    pinfo(INFO,0);
                }else{//only satisfy cif check
                    fprintf(fo, "#\n_diffrn_radiation_wavelength.id  1\n");
                    fprintf(fo, "_diffrn_radiation_wavelength.wavelength    .\n#\n");
                    
                    sprintf(INFO, "Warning: file (%s) has no _diffrn_radiation_wavelength.(given .)",outfile);
                    pinfo(INFO,0);
                }
                
            }
            
        }
        
    }
    if(line2 && nline2)free_memory_2d(line2, nline2);
    fclose(fo);
}

void usage(int n)
{
    printf("\n");
    printf("===================================================================\n");
    printf("           sf_convert (version: %s)                         \n",  VERSION );
    printf("===================================================================\n");
    printf("Usage: 'sf_convert  -i input_format -o output_format -sf data_file'\n");
    printf("   or simply type:  'sf_convert  -o output_format -sf data_file'\n");
    printf("===================================================================\n");
    printf("-i     input format: \n");
    printf("       mmCIF, CIF, MTZ, CNS, Xplor, HKL2000, Scalepack, Dtrek, TNT, \n");
    printf("       SHELX, SAINT, EPMR, XSCALE, XPREP, XTALVIEW, X-GEN, XENGEN, \n");
    printf("       MULTAN, MAIN, OTHER.\n");
    printf("-o     output format:\n");
    printf("       mmCIF, MTZ, CNS, TNT, SHELX, EPMR, XTALVIEW, HKL2000, Dtrek,\n");
    printf("       XSCALE, MULTAN, MAIN, OTHER.\n");
    printf("-sf    input structure factor file name.\n");
    printf("\nOptional options:\n");
    printf("-out   followed by output file name (If not given, default by program)\n");
    printf("-pdb   followed by PDB file (must contain symmetry and cell parameters)\n");
    printf("-cif   followed by cif file (must contain symmetry and cell parameters)\n");
    printf("-man   manually type in cells & symmetry (a,b,c,alpha,beta,gamma,p21)\n");
    printf("       (separate each item by a comma ',' without space).\n");
    printf("       This option can be replaced by -pdb!\n");
    printf("-flag  followed by a number for Rfree test set.\n");
    printf("       (e.g. '-flag 8' means 8%% of reflection selected for free R test)\n");
    printf("-sf_type  followed by I or F. (If not given, guessed by program)\n");
    printf("-label   followed by label name for CNS & MTZ (see 'sf_convert -h')\n");
    printf("-freer   followed by the free test value in CNS & MTZ (see 'sf_convert -h')\n");
    printf("-format  guess the format of the SF file (sf_convert -sf sffile -format)\n");
    printf("-reformat  reformat header part of sf file (sf_convert -reformat ciffile)\n");
    printf("-audit   update the audit record (ADD: -audit date :reason@ )\n");
    printf("    If only given '-audit date', add auto-corrected info. (yyyy-mm-dd)\n");
    printf("    If given '-audit  date :reason@', add reason (start ':' & end '@')\n");
    printf("-valid   (sf_convert -valid sffile) show various SF errors, and correct!\n");
    printf("-cif_check  Add this option to check pdbx dictionary. (local use only)\n");
    printf("-add_map   get map coefficients (FoFc & 2FoFc) adding this option.\n");
    printf("-rescut  (add '-rescut value', cut high resolution to value in SF_4_validate.cif\n");
    printf("-sigcut  (add '-sigcut value', cut I/SigI to value in SF_4_validate.cif\n");
    printf("-diags  (if add '-diags diagfile', export diagfile for warning/error message.");
    
    
    printf("\nExample: convert any supported formats to mmCIF:  \n");
    printf("    sf_convert  -o mmcif -sf sf_file_name\n");
    printf("\nExample: convert any supported formats to mtz:(cell in sf file)\n");
    printf("    sf_convert  -o MTZ  -sf sf_file \n"); 
    printf("\nExample: convert any supported formats to mtz:(cell NOT in sf file)\n");
    printf("    sf_convert  -o MTZ  -sf sf_file -pdb pdbfile\n"); 
    printf("    sf_convert  -o MTZ  -sf sf_file -man  a,b,c,alpha,beta,gamma,P21\n\n");
    printf("Default data type: F for EPMR, TNT, OTHER, MULTAN, MAIN,XTALVIEW;\n");
    printf("                   I for SHELX,HKL2000,DTREK,XSCALE,XPREP,SAINT;\n\n");
    printf("NOTE: If input SF file name is rxxxxsf.org, the xxxx is used as pdbid.\n\n");
    
    if(n==0) printf("Type 'sf_convert -h' for special conversion and detail instructions.\n\n");

    if(n==0) exit(0);
    

    printf("==============================================================================\n");
    printf("Note:  \n");
    printf(" 1. If MTZ format is involved, newer version of CCP4 must be installed!\n");
    printf(" 2. -pdb or -cif or -man must be given for conversion to MTZ file,\n");
    printf("    if input SF file does not contain symmetry and cells.\n");
    printf(" 3. If -man is used, separate each item by a comma ',' without space.\n");
    printf(" 4. If input format is OTHER, you should provide a ASCII file \n");
    printf("    with H, K, L, F, SigmaF separated by space.\n");
    printf(" 5. CIF and mmCIF are for small molecule and macro-molecule format.\n");

    printf("==============================================================================\n\n\
For CNS/Xplor/MTZ to mmCIF, if automatic conversion failed,\n\
or there are multiple datasets in the SF file, use option '-label'\n\
Each label be separated by a comma ',' and each data set be separated by a ':'.\n\
Default freeR set (1 for CNS; 0 for mtz). If not, specify by '-freer ?'. \n\n\
All the data labels follow those of CCP4 as below. \n\
THE LABELS ARE CASE SENSITIVE!!. \n\
\nIf the labels involved in the '(' or ')', they must be quoted by ' '\n\
\nExamples for manual conversion:\n\
 1. convert CNS to mmcif for one data set (with free set as 1)\n\
    sf_convert -i cns -o mmcif -sf data_file   \n\
    -label FP=? , SIGFP=?, FREE=? -freer 1 \n\n\
 2. convert CNS to mmcif for two data sets\n\
    sf_convert -i cns -o mmcif -sf data_file -label \n\
          FP=? ,SIGFP=?, I=?, SIGI=? , FREE=? : FP=? , SIGFP=?, I=?, SIGI=? \n\n\
 3. convert MTZ to mmcif for one data set\n\
    sf_convert -i mtz -o mmcif -sf data_file -label FP=? ,SIGFP=?, FREE=?\n\n\
 4. convert MTZ to mmcif for two data set\n\
    sf_convert -i mtz -o mmcif -sf data_file -label FP=? ,SIGFP=?,I=?,SIGI=?\n\
              FREE=? : FP=? , SIGFP=?, I=?, SIGI=? \n\n\
 5. convert MTZ to mmcif (one data set with anomalous)\n\
    sf_convert -i mtz -o mmcif -sf data_file -label FP=? ,SIGFP=? , \n\
              FREE=? , 'F(+)=?' ,'SIGF(+)=?','F(-)=?' ,'SIGF(-)=?'\n\n\
 6. convert MTZ to mmcif (specify free set to 2)\n\
    sf_convert -i mtz -o mmcif -sf data_file -freer 2 -label FP=? ,SIGFP=? , \n\
              FREE=? , 'F(+)=?' ,'SIGF(+)=?','F(-)=?' ,'SIGF(-)=?'\n\n\
Note: The question marks '?' correspond to the labels in the SF file.\n\
\n\
      mmCIF token                    type     data label   \n\
\n\
 _refln.F_meas_au         	      F           FP \n\
 _refln.F_meas_sigma_au               Q           SIGFP \n\
 _refln.intensity_meas    	      J           I     \n\
 _refln.intensity_sigma   	      Q           SIGI   \n\
\n\
 _refln.F_calc            	      F           FC     \n\
 _refln.phase_calc        	      P           PHIC   \n\
 _refln.phase_meas        	      P           PHIB   \n\
 _refln.fom               	      W           FOM    \n\
 \n\
_refln.pdbx_FWT                       F    \n\
_refln.pdbx_PHWT                      P    \n\
_refln.pdbx_DELFWT                    F    \n\
_refln.pdbx_DELPHWT                   P    \n\
 \n\
 _refln.pdbx_HL_A_iso                 A       	  HLA   \n\
 _refln.pdbx_HL_B_iso                 A       	  HLB   \n\
 _refln.pdbx_HL_C_iso                 A       	  HLC   \n\
 _refln.pdbx_HL_D_iso                 A       	  HLD   \n\
 \n\
 _refln.pdbx_F_plus         	      G           F(+)   \n\
 _refln.pdbx_F_plus_sigma   	      L           SIGF(+) \n\
 _refln.pdbx_F_minus        	      G           F(-)   \n\
 _refln.pdbx_F_minus_sigma  	      L           SIGF(-)\n\
 _refln.pdbx_anom_difference          D           DP   \n\
 _refln.pdbx_anom_difference_sigma    Q           SIGDP\n\
 _refln.pdbx_I_plus                   K           I(+)   \n\
 _refln.pdbx_I_plus_sigma             M           SIGI(+)\n\
 _refln.pdbx_I_minus                  K           I(-)   \n\
 _refln.pdbx_I_minus_sigma            M           SIGI(-) \n\
\n\
_refln.status            	      I           FREE    \n\
\
");
    
    exit(0);
}

int is_mtz(char *inpfile)
{
    char **line;
    int i, nline=0, mtz=0;

    line=get_lines_from_file(inpfile, &nline);
        
    for(i=0; i<nline; i++){
        if(line[0][0] =='M' && line[0][1] =='T' && line[0][2] =='Z' && line[0][3] ==' ' ) {
            mtz=1;
        }else
            break;
    }
    return mtz;
    
    if(nline)free_memory_2d(line, nline);
}
        
    
void guess_sf_format(char *inpfile, char *format)
// guess various format
{
    char **line,outfile[256],str[512],table[256],item[256];
    char s1[80],s2[80],s3[80],s4[80],s5[80],s6[80];
    int i, nline,  n=0, n1, n2,n3,n4,n5,n6,n7,n8,n9;
    int nn7=0,kk7=0, k1,k2,k3, m1,m2, mm1=0;
    float a;
    
    FILE *fout=NULL;

    
    line=get_lines_from_file(inpfile, &nline);
    if(nline<2){
        sprintf(INFO, "Error! File (%s) is wrong! too few lines (%d)\n", inpfile, nline);
        pinfo(INFO,0);
        exit(0);
    }
    
    strcpy(outfile, "sf_format_guess.text");
    fout=fopen(outfile, "w");
    
    n1=0; //guess mmCIF
    n2=0; //guess CIF
    n3=0; //guess CNS (MAIN)
    n4=0; //guess TNT
    n5=0; //guess XSCALE based on head
    n6=0; //guess d*TREK based on head
    n7=0; //guess HKL/SCALEPACK based on head or columns
    n8=0; //guess shelx
    n9=0; //guess saint
    k1=0;k2=0;k3=0;
    m1=0;m2=0;
    
    for(i=0; i<nline; i++){
        if(i==0 && line[i][0] =='M' && line[i][1] =='T' && line[i][2] =='Z' && line[i][3] ==' ' ){
            write_format(fout,"MTZ", 10);
            strcpy(format, "MTZ");
            return;
        }
        
        n=sscanf(line[i], "%s%s%s%s%s%s", str, s2, s3,s4, s5,s6);
        rid_of_end_space(line[i]);
        
            
        if(str[0]=='_' && strchr(str, '.') ){
            if (check_category(str, table, item)){
                if (!strcmp_case(table, "_refln")){
                    if (!strcmp_case(item, "F_meas_au") || !strcmp_case(item, "F_meas")){
                        k1=1;
                    }else if (!strcmp_case(item, "intensity_meas") || !strcmp_case(item, "F_squared_meas")){
                        k2=1;
                    }else if (!strcmp_case(item, "pdbx_F_plus") || !strcmp_case(item, "pdbx_I_plus")){
                        k3=1;
                    }
                }
                if (n1++>200 || k1 || k2 || k3) break;
                
            }else if(!strncmp_case(str, "_refln_", 7)){
                if(!strcmp_case(str, "_refln_F_meas" ))k1=1;
                if(!strcmp_case(str, "_refln_F_squared_meas" ))k2=1;
                
                if (n2++>200 || k1 || k2 ) break;
            }
            
        }else if(!strncmp_case(str, "INDE", 4) ||
                 (!strncmp_case(str, "DECLare",7)&& strstr_case(line[i]+20, "RECIproca"))){
            
            if(strstr_case(line[i], "FOBS=" ) || strstr_case(line[i], "FO=" )||
               strstr_case(line[i], " F_" ) || strstr_case(line[i], "F=" ))m1++;
            
            if(strstr_case(line[i], "IOBS=" ) ||  strstr_case(line[i], "IO=" )||
               strstr_case(line[i], "I=" ))m2++;

            if(strstr_case(line[i], " FOBS " )  )mm1++;
            
            if (n3++ >300) break;
            
            
        }else if(strstr_case(line[i], "FOBS=" ) || strstr_case(line[i], "FO=" )||
                 strstr_case(line[i], "F=" )){
            
            n3++;
            if (m1++ >300) break;
            
        }else if(strstr_case(line[i], "IOBS=" ) || strstr_case(line[i], " IO=" )||
                 strstr_case(line[i], " I=" )){
            n3++;
            if (m2++ >300) break;
            
        }else if(!strncmp_case(str, "HKL", 3)){  //TNT format
            sscanf(line[i], "%*s%*s%*s%*s%f", &a);
            if(a<0)m2++;
            if (n4++ >200) break;
            
        }else if(!strncmp_case(str, "!SPACE_GROUP_NUMBER=", 20) ||  //xscale
                 !strncmp_case(str, "!UNIT_CELL_CONSTANTS=", 21) ||
                 !strncmp_case(str, "!ITEM_H=",8) ||
                 !strncmp_case(str, "!ITEM_K=",8) ||
                 !strncmp_case(str, "!ITEM_L=",8) ){
            if (++n5>4) break;
            
        }else if(!strncmp_case(str, "CRYSTAL_MOSAICITY=", 18) ||  
                 !strncmp_case(str, "CRYSTAL_SPACEGROUP=",19) ||
                 !strncmp_case(str, "CRYSTAL_UNIT_CELL=",18)  ||
                 !strncmp_case(str, "nH",2) ||
                 !strncmp_case(str, "nK",2) ||
                 !strncmp_case(str, "nL",2) ){
            
            if (++n6>5) break;
            
        }else if(((i==0 && !strcmp(str, "1") && n==1) ||
                 (i==1 && (!strcmp(str, "-985") || !strcmp(str, "-987")) && n==1) ||
		  (i==2 && strchr(str,'.') && strlen(line[i])>60)) && n==6){
            if (++n7>=3) break;  //scalepack
            
            
        }else if(i>50 && strlen(line[i])==32 ){
            
            strncpy(s1, line[i]+12, 8);
            s1[8] = '\0';
            if(atof(s1)<0) m2++; //I
            n8++;
            
        }else if(i>50 && strlen(line[i])>150 ){
            if (++n9>5) break;
        }
        
        if (i>50){ //HKL or shelx (if strlen==28)
            if (strlen(line[i])==28 || strlen(line[i])==35 ||
                strlen(line[i])==44 || strlen(line[i])==53){
                nn7++;
                
            }
            if (kk7++ >2000) break;
        }
        
        
    }
//    printf("numbers %d %d %d %d %d %d %d %d %d : %d %d \n", n1,n2,n3,n4,n5,n6,n7, n8, n9, nn7, kk7);
    
    if(n1>=4){ 
        if(k1==1 && k2==1){
            write_format(fout,"MMCIF", 2); //I&F

        }else if(k1==1 && k2!=1){
            write_format(fout,"MMCIF", 1); //F
        }else if(k1!=1 && k2==1){
            write_format(fout,"MMCIF", 0); //I
        }else if(k3==1){
            write_format(fout,"MMCIF", 3); //ano
        }else
            write_format(fout,"MMCIF", -1); //nothing

        strcpy(format, "MMCIF");
        
    }else if(n2>=4){
        if(k1==1 && k2==1){
            write_format(fout,"CIF", 2); //I&F
        }else if(k1==1 && k2!=1){
            write_format(fout,"CIF", 1); //F
        }else if(k1!=1 && k2==1){
            write_format(fout,"CIF", 0); //I
        }else
            write_format(fout,"CIF", -1); //nothing
        
        strcpy(format, "CIF");
        
    }else if(n3>=50 && mm1<10){
        if(m1>100 && m2>100){
            write_format(fout,"CNS", 2); //I&F
        }else if(m1>50 &&  m2<10){
            write_format(fout,"CNS", 1); //F
        }else if(m2>50 &&  m1<10){
            write_format(fout,"CNS", 0); //I
            
        }else
            write_format(fout,"CNS", -1); //nothing

        strcpy(format, "CNS");
        
    }else if(n3>=50 && mm1>10){
        write_format(fout,"xplor", 1);
        strcpy(format, "xplor");
        
    }else if(n4>=100){
        if(m2>0)
            write_format(fout,"TNT", 0);
        else
            write_format(fout,"TNT", -2);
        
        strcpy(format, "TNT");
        
    }else if(n5>=5){
        write_format(fout,"XSCALE", 0);
        strcpy(format, "XSCALE");
        
    }else if(n6>=6){
        write_format(fout,"DTREK", 0);
        strcpy(format, "DTREK");
        
    }else if(n7>1 || (nn7>=200 && nn7==kk7)){
        write_format(fout,"HKL", 0);
        strcpy(format, "HKL");

    }else if(n8>=200){
        if(m2>0)
            write_format(fout,"SHELX", 0);
        else
            write_format(fout,"SHELX", -2);
        strcpy(format, "SHELX");
        
    }else if(n9>=200){
        write_format(fout,"SAINT", 0);
        strcpy(format, "SAINT");
        
    }else{
        write_format(fout,"", -1);
    }
    if(nline)free_memory_2d(line, nline);
        
    
}

void pinfo(char *info, int id)
//write all the log infor. about the SF file
{
    if (strstr_case(info, "Warning") || strstr_case(info, "Error")){
        fprintf(FTMP1, "%s\n",info);
        printf("%s\n",info);
    }else{
        if (id==0){
            fprintf(FTMP2, "%s\n", info);
            printf("%s\n",info);
        }else if (id==1){
            fprintf(FTMP2, "%s\n",info);
        }else if (id==2){
            printf("%s\n", info);
        }
    }
    
}

void  pinfo_local(FILE *fout, const char *format)
{
    printf("%s\n", format);
    fprintf(fout, "%s\n", format);
}

void  write_format(FILE *fout, const char *format, int key)
{
    int len=0;

    len=strlen(format);
    if(len<1){
        pinfo_local(fout, "Sorry! format can not be predicted. Check input file!");
        
    }else{
        printf("Input file FORMAT=%s\n", format);
        fprintf(fout,"FORMAT=%s\n", format);
        
        if(key==-1){
            pinfo_local(fout,"Data type can not be predicted. Check input file!");
        }
        
        if(key==-2){
            pinfo_local(fout,"Data type may be Amplitude(F), Check file to confirm!");
        }
        
        if(key==0){
            pinfo_local(fout, "Data type is Intensity (I).");
        }
        
        if(key==1) {
            pinfo_local(fout, "Data type is Amplitude (F).");
        }
        
        if(key==2) {
            pinfo_local(fout,"Data type are both Amplitude (F) and Intensity (I).");
        }
        
        if(key==3) {
            pinfo_local(fout,"Data type are anomolous signal.");
        }
        
        if(key==10) {
            pinfo_local(fout,"Data type for MTZ file can be auto_converted.");
        }
        
    }
    fclose(fout);
    
}


void get_cell_pdb(char *pdbfile, char *pdb_id)
{
    char str[100], symmetry_in[20], tmp[80];    
    FILE *fpdb;
    
    if((fpdb=fopen(pdbfile, "r"))==NULL){
        open_file_error(pdbfile, "get_cell_pdb");
    }
    
    while(fgets(str, sizeof str, fpdb)!=NULL){
        if (!strncmp(str,"HEADER ", 7)){
            get_string(str,61,66, tmp);
            rid_of_space(tmp);
            if(strlen(tmp)==4){
                lowerstr(tmp);
                strcpy(pdb_id, tmp);
            }
            
        }else if(strstr(str,"REMARK 200  WAVELENGTH OR RANGE        (A) :")){
            sscanf(strchr(str, ':')+1,"%s", tmp);
            if(isdigit(tmp[0])) {
                WAVE=atof(tmp);
                if(WAVE>2.0 || (WAVE<0.6 && WAVE>0.01)){
                    sprintf(INFO, "Warning:  wavelength (%.4f) in coord. is abnormal (double check)!", WAVE);
                    pinfo(INFO, 0);
                }
            }
            

        }else if (!strncmp(str,"CRYST1", 6)){
            sscanf(str+6, "%f%f %f%f %f%f",
                   &CELL[0], &CELL[1], &CELL[2], &CELL[3], &CELL[4], &CELL[5]);
            
            get_string(str,54,64, symmetry_in);
            space_group_convertor(symmetry_in, SYMM);
            break;
        }
    }
    fclose(fpdb);

}

void get_cell_man(char **argv, int argc, int in, int *out)
//manually input cell and symmetry
{
    char tmp[256], **line;
    int i, nstr;
    
    if(argv[in+1][0] =='-'){
        printf("Wrong input! It should be '-man  a,b,c,alpha,beta,gamma,SYMM' (no space)\n");
        *out=in+1;
        return;
    }
    
    strcpy(tmp,argv[in+1]);
    
    for(i=in+2; i<argc; i++){
        if(argv[i][0]=='-') break;strcat(tmp," , ");
        strcat(tmp,argv[i]);
    }
    rid_of_space(tmp);
    
    *out=i-1;
    
    line = string_token(tmp,",", &nstr);
    if(nstr!=7){
        printf("Wrong input! Number of input (%d) < 7\n", nstr);
        printf("The format is '-man  a,b,c,alpha,beta,gamma,P21'\n");
        return;
    }
    CELL[0]=atof(line[0]);
    CELL[1]=atof(line[1]);
    CELL[2]=atof(line[2]);
    CELL[3]=atof(line[3]);
    CELL[4]=atof(line[4]);
    CELL[5]=atof(line[5]);
    
    strcpy(SYMM, line[6]);
    if(line) free_memory_2d(line, nstr);
    

}



void check_input_error(int argc, char **argv)
{
    int  i;
/*    
    const char *option[]=
        {"-i","-o", "-sf", "-out", "-pdb", "-cif",  "-man", "-flag", "-sf_type",
          "-pdb_id","-format", "-mtzdmp", "-label", "-freer",  "END"
        };
*/    
    if(argc==1) {
        usage(0);
        exit(0);
    }else if(argc==2 && !strcmp_case(argv[1], "-h")){
        usage(1);
        exit(0);
    }else if(argc<2){
        printf("\nInput error! There are not enough arguments.\n");
        usage(0);
        exit(0);
    }
    for(i=1; i<argc; i++){
        if(!strcmp_case(argv[i],"MTZ")) {
            if((CCP4 = getenv("CCP4")) == NULL) {
                printf("Error! you did not set CCP4 environment!\n");
                return;
            }
            break;
        }
    }


}

void reformat_mtz_output(char *inpfile, char *pdb_id)
{
    char  **line,str[256], phs[70];
    FILE  *fout=NULL;
    int i,j,nline, m=0;

    strcpy(phs, "_ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789");
    line=get_lines_from_file(inpfile, &nline);

    
    for(i=0; i<nline; i++){
        rid_of_front_space(line[i]);
        if(line[i][0]=='#'){
           strcpy(line[i],"#\n");
 	}
        if (strstr(line[i], "_refln.pdbx_F_meas_plus")){//correct token in mtz2cif
            replace_first_str(line[i],"_F_meas_plus", "_F_plus", line[i]);
        }else if (strstr(line[i], "_refln.pdbx_F_meas_minus")){
            replace_first_str(line[i],"_F_meas_minus", "_F_minus", line[i]);
        }
    }
    
    fout = fopen(inpfile,"w");
    for(i=0; i<nline; i++){    
        if(!strncmp(line[i],"data_", 5)){
            
            if(!strcmp_case(pdb_id, "xxxx")){
                fprintf(fout, "%s", line[i]);
            } else{
                if(m==0){
                    fprintf(fout, "data_r%ssf\n", pdb_id);
                }else{
                    fprintf(fout, "data_r%s%csf\n", pdb_id, phs[m]);
                }
            }
            m++;
            
               
               
            fprintf(fout,"#\n%s\n%s\n%s\n#\n",
                    "_audit.revision_id      1_0 ",
                    "_audit.creation_date  ?",
                    "_audit.update_record    'Initial release' "
                    );
	     continue;
        }else if( (i>0 && line[i][0]=='#' &&  line[i-1][0]=='#') ){
	     continue;
        }else if(!strncmp(line[i],"_audit.", 7)){
	     for(j=i; j<i+20; j++){

                 
		if(line[j][0]=='#' || !strncmp(line[j],"_cell.", 6)||
		!strncmp(line[j],"loop_", 5)){
                    i=j;
                    break;
                 }
		continue;
              }
        }
        if(strstr_space(line[i], "id     mtz2cif") ){
            replace_first_str(line[i],"mtz2cif", pdb_id, str);
            fprintf(fout, "%s", str);
        }else{
            fprintf(fout, "%s", line[i]);
        }
        
    }
    fprintf(fout,"#END OF REFLECTIONS\n");
    fclose(fout);
    if(nline>0)free_memory_2d(line, nline); 
}



void cifparse(char *iFile, char *oFile, int width)
// read & write
{
  string diags;
  CifFile * fobjR = NULL;

  
  CifParser* cifParserR = NULL;
  fobjR = new CifFile();
  cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
  cifParserR->Parse(iFile, diags);
  delete(cifParserR);

  if (!diags.empty()) printf(" Diagnostics [ %s] \n", diags.c_str() );
  
  fobjR->Write(oFile, true);
  fobjR->Close();
  if (fobjR) delete fobjR;
  
  if (!diags.empty())
  {
      char* logFileNameP = (char*)malloc(strlen(iFile) + strlen("-parser.log") +
        1);

      memset(logFileNameP, 0, strlen(iFile) + strlen("-parser.log") + 1);
      strcpy(logFileNameP, iFile);
      strcat(logFileNameP, "-parser.log");

//      delete_file(logFileNameP);

      free(logFileNameP);
  }
}

int check_mtz_label(char *file)
{
    char **line;
    int nline, i, n=1;

    line=get_lines_from_file(file, &nline);
    
    for(i=0; i<nline; i++){
        if(strstr(line[i], "LABIN ") && !strstr(line[i], "=")){
            n=0;
            break;
        }
    }
    if(nline) free_memory_2d(line, nline);
    return n;
}

void check_unwanted_cif_item(char *file)
{
    char str[512], tmp[512], command[512];
    int k;
    FILE *fp=NULL;

    static const char *items[]={
"_atom_sites.entry_id",
"_audit_author.name",
"_audit.creation_method",
"_audit.update_recor",
"_cell.CCP4_crystal_id",
"_cell.CCP4_wavelength_id",
"_cell.enry_id",
"_cell.entry",
"_cell_entry.id",
"_cell.enrty_id",
"_cell.entry.id",
"_cell.ndb_unique_axis",
"_cell.Z_PDB",
"_database_2.database_code",
"_database_2.database_id",
"_database.entry_id",
"_database.ndb_code_NDB",
"_database.ndb_code_PDB",
"_diffrn.detail",
"_diffrn.pdbx_crystal_id",
"_diffrn_radiation.id",
"_diffrn_radiation.pdbx_wavelength_list",
"_diffrn_radiation.type",
"_diffrn_radiation_wavelength.CCP4_crystal_id",
"_diffrn_radiation_wavelength.wt",
"_pdbx_powder_refln.d_spacing",
"_pdbx_powder_refln.F_squared_calc",
"_pdbx_powder_refln.F_squared_meas",
"_pdbx_powder_refln.gsas_i100_meas",
"_pdbx_powder_refln.index_h",
"_pdbx_powder_refln.index_k",
"_pdbx_powder_refln.index_l",
"_pdbx_powder_refln.observed_status",
"_pdbx_powder_refln.phase_calc",
"_pdbx_reflns_twin.crystal_id",
"_pdbx_reflns_twin.diffrn_id",
"_pdbx_reflns_twin.fraction",
"_pdbx_reflns_twin.mean_F_square_over_mean_F2",
"_pdbx_reflns_twin.mean_I2_over_mean_I_square",
"_pdbx_reflns_twin.operator",
"_pdbx_reflns_twin.type",
"_phasing_set_refln.crystal_id",
"_phasing_set_refln.F_meas_au",
"_phasing_set_refln.fom",
"_phasing_set_refln.index_h",
"_phasing_set_refln.index_k",
"_phasing_set_refln.index_l",
"_phasing_set_refln.phase_meas",
"_phasing_set_refln.scale_group_code",
"_phasing_set_refln.status",
"_phasing_set_refln.wavelength_id",
"_radiation.id",
"_refine.entry_id",
"_refine.ls_d_res_high",
"_refine.ls_d_res_low",
"_refln.fiber_coordinate",
"_refln.fiber_F_meas_au",
"_refln.fiber_layer",
"_refln.F_meas",
"_refln.F_meas_sigma",
"_refln.F_meas_sigma_uncorrected",
"_refln.F_meas_uncorrected",
"_refln.F_part_au",
"_refln.gsas_i100_meas",
"_refln.intensity_calc",
"_refln.intensity_meas_au",
"_refln.intensity_meas_sigma_au",
"_refln.intensity_meas_unknown",
"_refln.intensity_sigma_unknown",
"_refln.mean_path_length_tbar",
"_refln.observed_status",
"_refln.pdbx_cos_phase_calc",
"_refln.pdbx_F_backtransform_au",
"_refln.pdbx_fom_weighted_fmap",
"_refln.pdbx_gsas_i100_meas",
"_refln.pdbx_HLA",
"_refln.pdbx_HLB",
"_refln.pdbx_HLC",
"_refln.pdbx_HLD",
"_refln.pdbx_phase_backtransform",
"_refln.pdbx_phase_cycle",
"_refln.pdbx_sin_phase_calc",
"_refln.phase_meas_sigma",
"_refln.phase_part",
"_refln.sgx_fmap",
"_refln.sint_over_lambda",
"_refln.symmetry_multiplicity",
"_refln.wavelength",
"_refln.waveLEngth_id",
"_refln.weight",
"_reflns.CCP4_crystal_id",
"_reflns.CCP4_wavelength_id",
"_reflns.observed_criterion_sigma_I",
"_struct_keywords.entry_id",
"_struct_keywords.ndb_keywords",
"_struct_keywords.text",
"_symmetry.cell_setting",
"__symmetry.entry_id",
"_symmetry.int_tables_number",
"_symmetry.ndb_full_space_group_name_H-M",
"_symmetry.space_group_name_h-m",
"END"
    };
    
    sprintf(tmp, "%s__tmp", file);
    sprintf(command, "grep \"^_\" %s |awk '{print $1}' | sort |uniq >%s", file, tmp);
    system(command);

    fp=fopen(tmp, "r");
    while(fgets(str, sizeof str, fp)!=NULL){
        rid_of_front_end_space(str);
        for (k=0; k<5000; k++){
            if(!strcmp(items[k], "END")) break;
            if(!strcmp(items[k], str)){
                sprintf(INFO, "Warning: File (%s) has unwanted CIF item (%s)",file,str);
                pinfo(INFO,0);
                break;
            } 
        }
    }
    fclose(fp);
    if (VERB==0) delete_file(tmp);
}


